/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "oirSupport.h"
#include "fitsio.h"
#include "imaging_fdata.h"

/**********************************************************
 * Perform an operation on all frames of an IMAGING_FDATA 
 * table using a specified row of a second table
 * All tables must have the same structure (not checked).
 * Usage:
 *    FrameMath  infile operation in2file row outfile
 *        infile IMAGING_FDATA table to be operated on
 *        operation 'ADD', 'SUBTRACT', 'MULTIPLY', 'DIVIDE'
 *        in2file 2nd IMAGING_FDATA
 *        row     1-rel row number in in2file to use in operation
 *        outfile is IMAGING_FDATA table with result.
 ************************************************************* */

/* Local prototypes   */
void doAdd (imaging_detector_table_def *detector, imaging_fdata_row *rowin1, 
	    imaging_fdata_row *rowin2, imaging_fdata_row *rowout);

void doSub (imaging_detector_table_def *detector, imaging_fdata_row *rowin1, 
	    imaging_fdata_row *rowin2, imaging_fdata_row *rowout);

void doMul (imaging_detector_table_def *detector, imaging_fdata_row *rowin1, 
	    imaging_fdata_row *rowin2, imaging_fdata_row *rowout);

void doDiv (imaging_detector_table_def *detector, imaging_fdata_row *rowin1, 
	    imaging_fdata_row *rowin2, imaging_fdata_row *rowout);

int main (int argc, char** argv) {

#define MAXREGION 12 /* Maximum number of regions */
   char *infile, *in2file, *outfile, *operator;
   int nFiles, iop;
   int i,ierr;
   char outFileName[130];
   char *bang="!";
   fitsfile *inFile=NULL, *in2File=NULL;
   fitsfile *outFile = NULL;
   imaging_fdata_table_def *odata, *idata, *i2data;
   imaging_detector_table_def *DetectorDef;
   imaging_fdata_row *outData, *in2Data;
   long iRow, nRow;
   int row; /* Row number in in2file to use */

   ierr = 0;
/*
      Get  inputs
*/
   infile = argv[1];
   operator = argv[2];
   in2file = argv[3];
   sscanf(argv[4], "%i", &row);
   outfile = argv[5];

/* 
      Parse operator 
*/
   iop = -1;
   if (!strcmp(operator,"ADD"))      iop = 1;
   if (!strcmp(operator,"SUBTRACT")) iop = 2;
   if (!strcmp(operator,"MULTIPLY")) iop = 3;
   if (!strcmp(operator,"DIVIDE"))   iop = 4;
   if (iop<0) { /* unknown */
     printf("Unknown operator %s\n", operator);
     return 1;
   }

/*
      Open input files
*/
   i = fits_open_file(&inFile, infile, READONLY, &ierr) ;
   if (ierr != 0) {
     printf("cfitsio error %i opening file %s\n", ierr, infile);
     printf("%s \n",strerror(errno));
     table_print_error_stack();  /* Show any cfitsio error messages */
     return 1;
   }
/*
  Open read second iinput file row 
*/
   i = fits_open_file(&in2File, in2file, READONLY, &ierr) ;
   if (ierr != 0) {
     printf("cfitsio error %i opening file %s\n", ierr, in2file);
     printf("%s \n",strerror(errno));
     table_print_error_stack();  /* Show any cfitsio error messages */
     return 1;
   }
   i2data = open_imaging_fdata_table(in2File, &ierr);
   if (ierr !=0) {
     printf("error %i opening imaging data in file %s\n", ierr, in2file);
     table_print_error_stack();  /* Show any cfitsio error messages */
     return 1;
   }
   in2Data = make_imaging_fdata_row(i2data);
   iRow = row;
   i = read_imaging_fdata_table(in2File, i2data, in2Data, iRow, &ierr);
   if (ierr !=0) {
     printf("error %i reading imaging data line %i in file %s\n", 
	    ierr, iRow, in2file);
     table_print_error_stack();  /* Show any cfitsio error messages */
     return 1;
   }
   i= close_imaging_fdata_table(in2File, i2data, &ierr);

/*
      Get detector from infile
*/
   DetectorDef = read_imaging_detector_table(inFile, &ierr);
   if (ierr !=0) {
     printf("error %i opening detector table in %s\n", ierr, infile);
     return 1;
   }
   
/* 
      Open first input IMAGING_FDATA table
*/
   idata = open_imaging_fdata_table(inFile, &ierr);
   if (ierr !=0) {
     printf("error %i opening imaging data in file %s\n", ierr, infile);
     table_print_error_stack();  /* Show any cfitsio error messages */
     return 1;
   }

/*
      Prepare output file
*/
   strcpy(outFileName, bang);
   strcat(outFileName, outfile);

   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error opening fits output %s %i\n", outfile, ierr);
      printf("%s \n",strerror(errno));
      table_print_error_stack();  /* Show any cfitsio error messages */
      return 1;
   }

   odata = make_imaging_fdata_table_def (outFile, DetectorDef, 1, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error creating output data definition\n");
      table_print_error_stack();  /* Show any cfitsio error messages */
      return 1;
   }

/*
     Create and initialize row, use same for input, output
*/
   outData = make_imaging_fdata_row(odata);

/* 
     Loop over input 
*/
   i = fits_get_num_rows (inFile,  &nRow, &ierr);
   for (iRow=0; iRow<nRow; iRow++) {
     /* Read */
     i = read_imaging_fdata_table(inFile, idata, outData, iRow, &ierr);
     if (ierr !=0) {
       printf("error %i reading imaging data line %i in file %s\n", 
	      ierr, iRow, infile);
       table_print_error_stack();  /* Show any cfitsio error messages */
       return 1;
     }
/*
      Do operation 
*/
     if (iop==1) { /* Add */
      doAdd (DetectorDef, outData, in2Data, outData);
     } else if (iop==2) { /* Subtract */
       doSub (DetectorDef, outData, in2Data, outData);
     } else if (iop==3) { /* Multiply */
       doMul (DetectorDef, outData, in2Data, outData);
     } else if (iop==4) { /* Divide */
       doDiv (DetectorDef, outData, in2Data, outData);
     } 

     /* Write */
     i = write_imaging_fdata_table(outFile, odata, outData, 1, &ierr);
     if (ierr !=0) {
       printf("error %i writing imaging data line %i in file %s\n", 
	      ierr, iRow, outfile);
       table_print_error_stack();  /* Show any cfitsio error messages */
       return 1;
     }
  } /* end loop over input file */

   /* Close */
   i = close_imaging_fdata_table(outFile, idata, &ierr);
   i = fits_close_file(inFile, &ierr);
   i = close_imaging_fdata_table(outFile, odata, &ierr);
   i = fits_close_file(outFile, &ierr);

/*
       Cleanup
*/
   kill_imaging_fdata_row(in2Data);
   kill_imaging_fdata_row(outData);

   return 0;
} /* end main */

/***************************************************************
 * IMAGING_FDATA math Add
 * Add corresponding values in rowin2 to rowin1 and output to rowout
 * Inputs:
 *    detector IMAGING_DETECTOR structure describilg input and output
 *    rowin1   first operand
 *    rowin2   second operand
 *    rowout   result
 **************************************************************** */
void doAdd (imaging_detector_table_def *detector, imaging_fdata_row *rowin1, 
	    imaging_fdata_row *rowin2, imaging_fdata_row *rowout)
{
    int npix, ipix, nregion, ireg;

    /* Loop over regions */
    nregion = detector->nregion;
    for (ireg=0; ireg<nregion; ireg++) {
      /* loop over pixels */
      npix = detector->naxis[ireg][0] * detector->naxis[ireg][1];
      for (ipix=0; ipix<npix; ipix++) {
        rowout->data[ireg][ipix] = rowin1->data[ireg][ipix] + rowin2->data[ireg][ipix];
      }
    } /* end loop over regions */
} /* end doAdd */

/***************************************************************
 * IMAGING_FDATA math SUBTRACT
 * Subtract corresponding values in rowin2 from rowin1 and output to rowout
 * Inputs:
 *    detector IMAGING_DETECTOR structure describing input and output
 *    rowin1   first operand
 *    rowin2   second operand
 *    rowout   result
 **************************************************************** */
void doSub (imaging_detector_table_def *detector, imaging_fdata_row *rowin1, 
	    imaging_fdata_row *rowin2, imaging_fdata_row *rowout)
{
    int npix, ipix, nregion, ireg;

    /* Loop over regions */
    nregion = detector->nregion;
    for (ireg=0; ireg<nregion; ireg++) {
      /* loop over pixels */
      npix = detector->naxis[ireg][0] * detector->naxis[ireg][1];
      for (ipix=0; ipix<npix; ipix++) {
        rowout->data[ireg][ipix] = rowin1->data[ireg][ipix] - rowin2->data[ireg][ipix];
      }
    } /* end loop over regions */
} /* end doSub */

/***************************************************************
 * IMAGING_FDATA math MULTIPLY
 * Multiply corresponding values in rowin2 by rowin1 and output to rowout
 * Inputs:
 *    detector IMAGING_DETECTOR structure describilg input and output
 *    rowin1   first operand
 *    rowin2   second operand
 *    rowout   result
 **************************************************************** */
void doMul (imaging_detector_table_def *detector, imaging_fdata_row *rowin1, 
	    imaging_fdata_row *rowin2, imaging_fdata_row *rowout)
{
    int npix, ipix, nregion, ireg;

    /* Loop over regions */
    nregion = detector->nregion;
    for (ireg=0; ireg<nregion; ireg++) {
      /* loop over pixels */
      npix = detector->naxis[ireg][0] * detector->naxis[ireg][1];
      for (ipix=0; ipix<npix; ipix++) {
        rowout->data[ireg][ipix] = rowin1->data[ireg][ipix] * rowin2->data[ireg][ipix];
      }
    } /* end loop over regions */
} /* end doMul */

/***************************************************************
 * IMAGING_FDATA math DIVIDE
 * Divide corresponding values in rowin2 into rowin1 and output to rowout
 * Inputs:
 *    detector IMAGING_DETECTOR structure describilg input and output
 *    rowin1   first operand
 *    rowin2   second operand
 *    rowout   result
 **************************************************************** */
void doDiv (imaging_detector_table_def *detector, imaging_fdata_row *rowin1, 
	    imaging_fdata_row *rowin2, imaging_fdata_row *rowout)
{
    int npix, ipix, nregion, ireg;

    /* Loop over regions */
    nregion = detector->nregion;
    for (ireg=0; ireg<nregion; ireg++) {
      /* loop over pixels */
      npix = detector->naxis[ireg][0] * detector->naxis[ireg][1];
      for (ipix=0; ipix<npix; ipix++) {
        rowout->data[ireg][ipix] = rowin1->data[ireg][ipix] / rowin2->data[ireg][ipix];
      }
    } /* end loop over regions */
} /* end doDiv */
